#include <stdio.h>
#include <stdlib.h>

typedef struct data_t {
    long id;
    void *data;
} data_t;

// struttura node dell'albero
typedef struct node_t {
    struct data_t *data;
    struct node_t *left; 
    struct node_t *right;
} node_t;

// costruisce un albero binario di ricerca 
struct node_t *buildTree(data_t *elem, node_t *t) {
    if (t == NULL) { // albero vuoto
	t = (node_t*)malloc(sizeof(node_t));
	if (!t) {
	    perror("malloc");
	    return NULL;
	}
	t->data  = elem;	
	t->left  = NULL;
	t->right = NULL;
    } else {
	if (elem->id < t->data->id) // elemento minore della radice
	    t->left  = buildTree(elem, t->left); // percorro il sottoalbero sinistro 
	else 
	    t->right = buildTree(elem, t->right); // percorro il sottoalbero destro
    }
    return t;
}

// dealloco l'albero
void deleteTree(node_t *t, void (*F)(void*)) {
    if (t != NULL) {
	deleteTree(t->left, F);
	deleteTree(t->right, F);
	if (F) F(t->data);
	free(t);
    }
}

// si suppone che t != NULL, il controllo va fatto prima
// della chiamata a getMin
data_t *getMin(node_t *t) {
    if (t->left==NULL)  // albero di sinistra vuoto
	return t->data;	// il minimo e' la radice
    else
	return getMin(t->left);	
}

// si suppone che t != NULL, il controllo va fatto prima
// della chiamata a getMax
data_t *getMax(node_t *t) {
    if (t->right == NULL) // albero di destra vuoto
	return t->data;   // il massimo e' la radice
    else 
	return getMax(t->right);
}
// stampa i nodi dell'albero facendo una visita ordinata
void printInOrder(node_t *t) {
    if (t != NULL) {
	printInOrder(t->left);
	printf("%ld ", t->data->id);
	printInOrder(t->right);
    }
}

data_t *searchId(long id, node_t *root) {
    node_t *n = root;
    while(n != NULL && n->data->id != id) {
	if (id < n->data->id) n = n->left;
	else n = n->right;
    }
    return n?n->data:NULL;
}


data_t *removeId(long id, node_t **root, void (*F)(void*)) {
    if (*root == NULL) return NULL;

    if ((*root)->data->id == id) {
	if ((*root)->left == NULL && (*root)->right == NULL) {
	    data_t *d = (*root)->data;
	    if (F) F(*root);
	    *root = NULL;
	    return d;
	}
	if ((*root)->right == NULL) {
	    data_t *d = (*root)->data;
	    node_t *tmp = (*root)->left;
	    if (F) F(*root);
	    *root = tmp;
	    return d;
	} else if ((*root)->left == NULL) {
	    data_t *d = (*root)->data;
	    node_t *tmp = (*root)->right;
	    if (F) F(*root);
	    *root = tmp;
	    return d;
	}
	/* il minimo di destra sostituisce il nodo corrente */

	data_t *min   = getMin((*root)->right);
	int   oldid   =	(*root)->data->id;
	void *olddata = (*root)->data->data;
	(*root)->data->id   = min->id;
	(*root)->data->data = min->data;
	data_t *d = removeId(min->id, &(*root)->right, F); // rimuoviamo il duplicato che si e' formato
	d->id   = oldid;
	d->data = olddata;
	return d;	    
    }
    data_t *d = NULL;
    if (id < (*root)->data->id)
	d = removeId(id, &(*root)->left, F);
    else 
	d = removeId(id, &(*root)->right, F);
    return d;
}

int main(int argc, char *argv[]) {
    if (argc == 1) {
	fprintf(stderr, "usa: %s id [ids]\n", argv[0]);
	return -1;
    }
    node_t *root = NULL;
    const long array_size = 10;
    data_t array[] = { {12, "ciao"}, {32, "mondo"}, {18,"hello"}, {-1,"pippo"}, 
		       {0, "paperino"}, {18, "qua"}, {5, "quo"}, {54, "pluto"}, {28, "topolino"}, {15,"qui"}  };
    
    for(long i=0; i<array_size; ++i)
	root = buildTree(&array[i], root);

    printf("Min: %ld\n", (getMin(root))->id);
    printf("Max: %ld\n", (getMax(root))->id);

    printInOrder(root);
    printf("\n");

    for(int i=1;i<argc;++i) {
	data_t *d = searchId(atoi(argv[i]), root);
	if (d!=NULL) {
	    printf("%ld trovato %s\n", d->id, (char *)d->data);
	} else {
	    printf("%s non trovato\n", argv[i]);
	}
    }

    for(int i=1;i<argc;++i) {
	data_t *d = removeId(atoi(argv[i]), &root, free);
	if (d) {
	    printf("%ld %s rimosso\n", d->id, (char*)(d->data));
	}
    }
    
    
    printInOrder(root);
    printf("\n");

    deleteTree(root, NULL);

    return 0;
}
